#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <time.h>
#include <malloc.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <ctype.h>

#include "util.h"
#include "object.h"
#include "classmember.h"
#include "directory.h"
#include "archive.h"
#include "source.h"
#include "collection.h"
#include "file.h"
#include "class.h"
#include "variable.h"
#include "function.h"
#include "edge.h"
#include "asxo/asxo.h"
#include "cx/cx.h"

// #define SANITY

static char	*cwdP = 0;
static char *g_program_nameP = 0;
static int	g_ret            = 0;

static char	*asx_compileP = 0;
static char *asx_unlinkP  = 0;
static char *asx_ignoreP  = 0;
static char *asx_forceP   = 0;
static char *asx_silentP  = 0;
static int  compile_flag  = 1;
static int	unlink_flag   = 1;
static int	silent_flag   = 1;
static int  use_stdin     = 0;
static int	lineno        = 0;

static int	show_variables   = Global_variables;
static int	no_function_calls= 0;

#define SUFFIX_a    0
#define SUFFIX_so   1
#define SUFFIX_o 	2
#define SUFFIX_s 	3
#define SUFFIX_S    4
#define SUFFIX_c 	5
#define SUFFIX_C 	6
#define SUFFIX_cc 	7
#define SUFFIX_cxx	8
#define SUFFIX_cpp	9
#define SUFFIX_cpp1 10
#define SUFFIX_none 11

/* Must agree with SUFFIX_* */

static char *suffixes[] =
{
	"a",
	"so",
	"o",
	"s",
	"S",
	"c",
	"C",
	"cc",
	"cxx",
	"cpp",
	"c++",
	0
};

static int ignore[sizeof(suffixes)/sizeof(*suffixes)] = {0};

static int force[sizeof(suffixes)/sizeof(*suffixes)] = {0};


static char	label_char[256]    = {0};
static char	function_char[256] = {0};

/* Convert the pathP to an absolute path.
 */

static char *
absolute_path(char *pathP)
{
	static char	full_path[2048];

	char	*P, *P1;

	P = full_path;
	if (*pathP == '/') {
		/* If the path is an absolute path we are done */
		strcpy(full_path, pathP);
	} else if (!cwdP || cwdP[0] != '/') {
		/* Don't know what the cwd is */
		strcpy(P, pathP);
	} else {
		strcpy(P, cwdP);
		P += strlen(cwdP);
		*P++ = '/';
		strcpy(P, pathP);
	}

	for (P = full_path; P1 = strchr(P, '.'); ) {
		switch (P1[1]) {
		case '/':
			if (P1 == full_path || P1[-1] == '/') {
				// Erase ./
				P = P1;
				strcpy(P, P+2);
				continue;
			}
			break;
		case '.':
			if (P1[2] == '/' && P1 > full_path+1 && P1[-1] == '/') {
				// Remove <parent>/../
				for (P = P1-2; P != pathP && P[-1] != '/'; --P);
				strcpy(P, P1+3);
				continue;
			}
			break;
		}
		P = P1 + 1;
	}
	return(full_path);
}

static int
call_cc(char *argv[], char * const env[], int suffix)
{
	int		ret, pid;
	int		status;
	int 	i;

	if (!silent_flag) {
		fprintf(stderr, "%4d: ", lineno);
		for (i = 0; argv[i]; ++i) {
			fprintf(stderr, "%s ", argv[i]);
		}
		fprintf(stderr,"\n");
	}
	
	ret = fork();
	switch (ret) {
	case 0:
		execve(argv[0], argv, env);
		fprintf(stderr,"** Line %d: Return from execve %s\n", lineno, argv[0]);
		error_exit(1);
	case -1:
		fprintf(stderr,"** Line %d: Fork to cc failed\n", lineno);
		perror("");
		error_exit(2);
	}
	pid = wait(&status);
	if (pid != ret) {
		fprintf(stderr,"** Line %d: Wait for cc failed\n", lineno);
		perror("");
		error_exit(3);
	}
	if (!WIFEXITED(status)) {
		fprintf(stderr,"** Line %d: cc did not terminate normally\n", lineno);
		return(-1);
	}
	return(WEXITSTATUS(status));
}

static void
dump_instances(void)
{
	Cdirectory	*directoryP;
	Carchive	*archiveP;
	Csource		*sourceP;
	Cclass		*classP;
	Cfunction	*functionP;
	Cvariable	*variableP;
	unsigned int	i, cnt;

	cnt = 0;
	for (directoryP = Cdirectory::g_headP; directoryP; directoryP = directoryP->m_nextP) {
		directoryP->emit_instance();
		++cnt;
	}
	if (cnt != Cdirectory::g_cnt) {
		fprintf(stderr, "** !!Warning!! - Lost directories %d seen but %d created\n", cnt, Cdirectory::g_cnt);
	}

	cnt = 0;
	for (archiveP = Carchive::g_headP; archiveP; archiveP = archiveP->m_nextP) {
		archiveP->emit_instance();
		++cnt;
	}
	if (cnt != Carchive::g_cnt) {
		fprintf(stderr, "** !!Warning!! - Lost archives %d seen but %d created\n", cnt, Carchive::g_cnt);
	}

	cnt = 0;
	for (sourceP = Csource::g_headP; sourceP; sourceP = sourceP->m_nextP) {
		sourceP->emit_instance();
		++cnt;
	}
	if (cnt != Csource::g_cnt) {
		fprintf(stderr, "** !!Warning!! - Lost sources %d seen but %d created\n", cnt, Csource::g_cnt);
	}

	cnt = 0;
	if ((classP = Cclass::g_headP)) {
		fputs("$INSTANCE C C\n", stdout);
		for (; classP; classP = classP->m_nextP) {
			classP->emit_instance();
			++cnt;
	}	}
	if (cnt != Cclass::g_cnt) {
		fprintf(stderr, "** !!Warning!! - Lost classes %d seen but %d created\n", cnt, Cclass::g_cnt);
	}

	cnt = 0;
	for (i = 0; i < HASHSIZE; ++i) {
		for (variableP = Cvariable::g_hash[i]; variableP; variableP = variableP->m_hashP) {
			if (!(variableP->m_flags & IgnoreX)) {
				variableP->emit_instance();
			}
			++cnt;
	}	}
	if (cnt != Cvariable::g_hash_cnt) {
		fprintf(stderr, "** !!Warning!! - Lost variables %d seen but %d created\n", cnt, Cvariable::g_hash_cnt);
	}

	cnt = 0;
	for (i = 0; i < HASHSIZE; ++i) {
		for (functionP = Cfunction::g_hash[i]; functionP; functionP = functionP->m_hashP) {
			functionP->emit_instance();
			++cnt;
	}	}
	if (cnt != Cfunction::g_hash_cnt) {
		fprintf(stderr, "** !!Warning!! - Lost functions %d seen but %d created\n", cnt, Cfunction::g_hash_cnt);
	}
}

static void
dump_edges(void)
{
	Cobject		*objectP;
	unsigned int cnt;

	cnt = 0;
	for (objectP = Cobject::g_created_headP; objectP; objectP = objectP->m_created_nextP) {
		objectP->emit_edges();
		++cnt;
	}
	if (cnt != Cobject::g_cnt) {
		fprintf(stderr, "** !!Warning!! - Lost objects %d seen but %d created\n", cnt, Cobject::g_cnt);
	}

}

static void
dump_attributes(void)
{
	Cdirectory	*directoryP;
	Carchive	*archiveP;
	Csource		*sourceP;
	Cclass		*classP;
	Cvariable	*variableP;
	Cfunction	*functionP;
	Cedge		*edgeP;
	unsigned int i;
	unsigned int cnt;

	printf("\nFACT ATTRIBUTE :\n\n");

	cnt = 0;
	for (directoryP = Cdirectory::g_headP; directoryP; directoryP = directoryP->m_nextP) {
		directoryP->emit_attributes();
		++cnt;
	}
	if (cnt != Cdirectory::g_cnt) {
		fprintf(stderr, "** !!Warning!! - Lost directories attributes %d seen but %d created\n", cnt, Cdirectory::g_cnt);
	}

	cnt = 0;
	for (archiveP = Carchive::g_headP; archiveP; archiveP = archiveP->m_nextP) {
		archiveP->emit_attributes();
		++cnt;
	}
	
	if (cnt != Carchive::g_cnt) {
		fprintf(stderr, "** !!Warning!! - Lost archives attributes %d seen but %d created\n", cnt, Carchive::g_cnt);
	}

	cnt = 0;
	for (sourceP = Csource::g_headP; sourceP; sourceP = sourceP->m_nextP) {
		sourceP->emit_attributes();
		++cnt;
	}
	if (cnt != Csource::g_cnt) {
		fprintf(stderr, "** !!Warning!! - Lost source attributes %d seen but %d created\n", cnt, Csource::g_cnt);
	}

	cnt = 0;
	if ((classP = Cclass::g_headP)) {
		printf("C { label=\":classes\" }\n");
		for (; classP; classP = classP->m_nextP) {
			classP->emit_attributes();
			++cnt;
	}	}
	if (cnt != Cclass::g_cnt) {
		fprintf(stderr, "** !!Warning!! - Lost class attributes %d seen but %d created\n", cnt, Cclass::g_cnt);
	}

	cnt = 0;
	for (i = 0; i < HASHSIZE; ++i) {
		for (variableP = Cvariable::g_hash[i]; variableP; variableP = variableP->m_hashP) {
			if (!(variableP->m_flags & IgnoreX)) {
				variableP->emit_attributes();
			}
			++cnt;
	}	}
	if (cnt != Cvariable::g_hash_cnt) {
		fprintf(stderr, "** !!Warning!! - Lost variable attributes %d seen but %d created\n", cnt, Cvariable::g_hash_cnt);
	}

	cnt = 0;
	for (i = 0; i < HASHSIZE; ++i) {
		for (functionP = Cfunction::g_hash[i]; functionP; functionP = functionP->m_hashP) {
			functionP->emit_attributes();
			++cnt;
	}	}

	if (cnt != Cfunction::g_hash_cnt) {
		fprintf(stderr, "** !!Warning!! - Lost function attributes %d seen but %d created\n", cnt, Cfunction::g_hash_cnt);
	}

	cnt = 0;
	for (i = 0; i < HASHSIZE; ++i) {
		for (edgeP = Cedge::g_hash[i]; edgeP; edgeP = edgeP->m_hashP) {
			edgeP->emit_attributes();
			++cnt;
	}	}
	if (cnt != Cedge::g_hash_cnt) {
		fprintf(stderr, "** !!Warning!! - Lost edge attributes %d seen but %d created\n", cnt, Cedge::g_hash_cnt);
	}

}

static void
dump_schema(void)
{
	fputs(				"\nSCHEME TUPLE :\n\n"
		  				"$INHERIT D   $ENTITY\n"	// Directory
						"$INHERIT A   D\n"			// Archive
						"$INHERIT S   $ENTITY\n"	// Source
						"$INHERIT C   S\n"			// Collection
						"$INHERIT F   $ENTITY\n"	// Function
						"$INHERIT GF  F\n"			// GlobalFunction
						"$INHERIT EF  GF\n"			// ExternalFunction
						"$INHERIT LF  F\n"			// LocalFunction
			,stdout);

	if (show_variables != No_variables || !no_function_calls) {
		fputs(			"$INHERIT V   $ENTITY\n"	// Variable
			,stdout);
	}

	if (show_variables != No_variables) {
		fputs(			"$INHERIT GV  V\n"			// GlobalVariable
						"$INHERIT WV  GV\n"			// WeakGlobalVariable
						"$INHERIT EV  GV\n"			// ExternalVariable
						"$INHERIT SV  V\n"			// StaticVariable
						"$INHERIT FV  SV\n"			// StaticFunctionVariable
			,stdout);
	}
	if (!no_function_calls) {
		fputs(          "$INHERIT FA  V\n"			// FunctionAddress
			,stdout);
	}
	if (Cclass::g_headP) {
		fputs(			"$INHERIT C   $ENTITY\n"	// Class
			,stdout);
	}

	fputs(				"contain  D   D\n" 		// Directory contains Directory
		  				"contain  D   S\n"		// Directory contains Source
		  				"contain  S   F\n"		// Source    contains Functions
						"contain  C   F\n"
		, stdout);

	if (show_variables != No_variables || !no_function_calls) {
		fputs(			"contain  S   V\n"		// Source    contains Variables
						"contain  C   V\n"
						"contain  F   V\n"		// Function  contains Variables
			, stdout);
	}
	if (!no_function_calls) {
		fputs(			"contain  F   FA\n"		// Function  contains address
						"CGF      F   GF\n"     // CallsGlobalFunction
						"CEF      F   EF\n"     // CallsExternalFunction
						"CLF      F   LF\n"		// CallsLocalFunction
						"CIF      F   FA\n"		// CallsIndirectFunction
			,stdout);
	}
	if (show_variables != No_variables) {
		fputs(			"RGV      F   GV\n"      // ReadGlobalVariable
						"REV      F   EV\n"      // ReadExternalVariable
						"RSV      F   SV\n"      // ReadStaticVariable
						"RFV      F   FV\n"		 // ReadStaticFunctionVariable
						"UGV      F   GV\n"      // UsesGlobalVariable
						"UEV      F   EV\n"      // UsesExternalVariable
						"USV      F   SV\n"      // UsesStaticVariable
						"UFV      F   FV\n"		 // UsesStaticFunctionVariable
			,stdout);
	}

	if (Cclass::g_headP) {
		fputs(			"M        C   F\n", stdout);	// Member of class is Function
		if (show_variables != No_variables) {
			fputs(		"M        C   V\n", stdout);	// Member of class is Variable
	}	}

	fputs(				"\nSCHEME ATTRIBUTE :\n\n"
		  				"$ENTITY {\n"
		  				"   color       = (255 255 255)\n"
          				"   labelcolor  = (  0   0   0)\n"
          				"   class_style = 1\n"
          				"   class_icon  = entity.png\n"
          				"   file\n"
		  				"}\n"
						"D {\n"
                        "   class_label = \"Directory\"\n"
		  				"   color       = (255 255 153)\n"
          				"   labelcolor  = (  0   0   0)\n"
          				"   class_icon  = directory.png\n"
          				"   class_style = 4\n"
		  				"}\n"
						"A {\n"
						"   class_label = \"Library\"\n"
		  				"   color       = (153 255 102)\n"
          				"   labelcolor  = (  0   0   0)\n"
          				"   class_icon  = library.png\n"
          				"   class_style = 4\n"
		  				"}\n"
						"S {\n"
						"   class_label = \"Source\"\n"
		  				"   color       = (  0 204   0)\n"
          				"   labelcolor  = (  0   0   0)\n"
          				"   class_icon  = source.png\n"
          				"   class_style = 0\n"
		  				"}\n"
						"C {\n"
						"   class_label = \"Collection\"\n"
		  				"   color       = (255 255 204)\n"
          				"   labelcolor  = (  0   0   0)\n"
          				"   class_icon  = collection.png\n"
          				"   class_style = 0\n"
		  				"}\n"
		  				"F {\n"
						"   class_label = \"Function\"\n"
          				"   class_style = 1\n"
          				"   class_icon  = function.png\n"
          				"   lineno\n"
		  				"}\n"
		  				"GF {\n"
						"   class_label = \"Global Function\"\n"
		  				"   color       = (255 102 102)\n"
          				"   class_icon  = globalfunction.png\n"
          				"   labelcolor  = (  0   0   0)\n"
          				"   class_style = 1\n"
		  				"}\n"
						"EF {\n"
						"   class_label = \"External Function\"\n"
		  				"   color       = (255   0 204)\n"
          				"   labelcolor  = (  0   0   0)\n"
          				"   class_style = 1\n"
          				"   class_icon  = externalfunction.png\n"
						"   class_description = \"Undeclared function\"\n"
		  				"}\n"
						"LF {\n"
						"   class_label = \"Local Function\"\n"
		  				"   color       = (255 153 153)\n"
          				"   labelcolor  = (  0   0   0)\n"
          				"   class_icon  = localfunction.png\n"
          				"   class_style = 1\n"
		  				"}\n", stdout);

	if (show_variables != No_variables || !no_function_calls) {
		fputs(			"V {\n"
						"   class_label = \"Variable\"\n"
          				"   class_icon  = variable.png\n"
                  		"   class_style = 6\n"
			      		"}\n"
			,stdout);
	}
	if (show_variables != No_variables) {
		fputs( 			"GV {\n"
						"  class_label = \"Global Variable\"\n"
                  		"  color       = (  0   0 255)\n"
                  		"  labelcolor  = (  0   0   0)\n"
          				"  class_icon  = globalvariable.png\n"
                  		"  class_style = 6\n"
			      		"}\n"
						"WV {\n"
						"  class_label = \"Weak Global Variable\"\n"
                  		"  color       = (204 204 255)\n"
                  		"  labelcolor  = (  0   0   0)\n"
          				"  class_icon  = weakvariable.png\n"
                  		"  class_style = 6\n"
			      		"}\n"
						"EV {\n"
						"  class_label = \"External Variable\"\n"
                  		"  color       = (153  51 255)\n"
                  		"  labelcolor  = (  0   0   0)\n"
          				"  class_icon  = externalvariable.png\n"
                  		"  class_style = 6\n"
						"  class_description = \"Undeclared variable\"\n"
			      		"}\n"
						"SV {\n"
						"  class_label = \"Static Variable\"\n"
                  		"  color       = (  0 204 204)\n"
                  		"  labelcolor  = (  0   0   0)\n"
          				"  class_icon  = staticvariable.png\n"
                  		"  class_style = 6\n"
						"  class_description = \"File scope variable\"\n"
			      		"}\n"
						"FV {\n"
						"  class_label = \"Static Function Variable\"\n"
                      	"  color       = (102 255 255)\n"
                      	"  labelcolor  = (  0   0   0)\n"
          				"  class_icon  = functionvariable.png\n"
                      	"  class_style = 6\n"
						"  class_description = \"Function scope variable\"\n"
			          	"}\n"
		,stdout);
	}
	if (!no_function_calls) {
		fputs(			"FA {\n"
						"  class_label = \"Function Address\"\n"
                      	"  color       = (102 255 128)\n"
                      	"  labelcolor  = (  0   0   0)\n"
          				"   class_icon  = functionaddress.png\n"
                      	"  class_style = 6\n"
						"  class_description = \"Address of function\"\n"
			          	"}\n", stdout);
	}

	if (Cclass::g_headP) {
		fputs(			"C {\n"
						"   class_label = \"Class\"\n"
		      			"   color       = (255 204   0)\n"
              			"   labelcolor  = (  0   0   0)\n"
          				"   class_icon  = class.png\n"
              			"   class_style = 1\n"
			          	"}\n", stdout);
	}

	if (show_variables != No_variables || !no_function_calls) {
		fputs(			"($RELATION) {\n"
						"   file\n"
						"   lineno\n"
						"   freq\n"
						"}\n"
			, stdout);
	}

	if (!no_function_calls) {
		fputs(			"(CGF) {\n"
						"   class_label = \"Calls Global Function\"\n"
						"   color       = (153   0   0)\n"
						"   class_style = 0\n"
						"}\n"
						"(CEF) {\n"
						"   class_label = \"Calls External Function\"\n"
						"   color       = (153   0   0)\n"
						"   class_style = 0\n"
						"}\n"
						"(CLF)  {\n"
						"   class_label = \"Calls Local Function\"\n"
						"   color       = (153   0   0)\n"
						"   class_style = 0\n"
						"}\n"
						"(CIF) {\n"
						"   class_label = \"Calls Function by address\"\n"
						"   color       = (255 0  0)\n"
						"   class_style = 0\n"
						"}\n" 
			,stdout);
	}
	if (show_variables != No_variables) {
		fputs(			"(RGV) {\n"
						"   class_label = \"Read Global Variable\"\n"
						"   color       = (  0 153   0)\n"
						"   class_style = 0\n"
						"}\n"
						"(REV) {\n"
						"   class_label = \"Read External Variable\"\n"
						"   color       = (  0 153   0)\n"
						"   class_style = 0\n"
						"}\n"
						"(RSV) {\n"
						"   class_label = \"Read Static Variable\"\n"
						"   color       = (  0 153   0)\n"
						"   class_style = 0\n"
						"}\n"
						"(RFV) {\n"
						"   class_label = \"Read Static Function Variable\"\n"
						"   color       = (  0 153   0)\n"
						"   class_style = 0\n"
						"}\n"
						"(UGV) {\n"
						"   class_label = \"Updates Global Variable\"\n"
						"   color       = (204   0 204)\n"
						"   class_style = 0\n"
						"}\n"
						"(UEV) {\n"
						"   class_label = \"Updates External Variable\"\n"
						"   color       = (204   0 204)\n"
						"   class_style = 0\n"
						"}\n"
						"(USV) {\n"
						"   class_label = \"Updates Static Variable\"\n"
						"   color       = (204   0 204)\n"
						"   class_style = 0\n"
						"}\n"
						"(UFV) {\n"
						"   class_label = \"Updates Static Function Variable\"\n"
						"   color       = (204   0 204)\n"
						"   class_style = 0\n"
						"}\n"
				, stdout);
	}

	if (Cclass::g_headP) {
		fputs(			"(M) {\n"
						"   class_label = \"Member\"\n"
						"   color       = (255 255 255)\n"
	    				"   class_style = 0\n"
						"}\n", stdout);
	}

	fputs(				"\nFACT TUPLE :\n\n", stdout);
}

static void
collapse_directories(void)
{
	Cdirectory	*directoryP, *parentP;
	Cdirectory	*collapseP;

	for (directoryP = Cdirectory::g_headP; directoryP; directoryP = directoryP->m_nextP) {
		if ((parentP = (Cdirectory *) directoryP->m_parentP)) {
			parentP->m_collapse++;
	}	}
	
	collapseP = 0;
	for (directoryP = Cdirectory::g_headP; directoryP; directoryP = directoryP->m_nextP) {
		parentP = (Cdirectory *) directoryP->m_parentP;
		if (!parentP) {
			// No point collapsing a directory that is already the root
			continue;
		}
		for (;; parentP = (Cdirectory *) parentP->m_parentP) {
			if (!parentP) {
				if (!collapseP) {
					collapseP = directoryP;
				}
				break;
			}
			if (parentP->m_collapse > 1) {
				break;
			}
			if (parentP == collapseP) {
				collapseP = directoryP;
	}	}	}
		
	if (collapseP) {

		char		buffer[2048];
		int			lth;

		lth = collapseP->full_path(buffer);
		buffer[--lth] = 0;
		collapseP->name(buffer);

		for (parentP = collapseP; parentP = (Cdirectory *) parentP->m_parentP; ) {
			parentP->m_collapse = -2;
		}
		collapseP->m_parentP = 0;
}	}

static void
dump_ta(void)
{
#ifdef SANITY
	void sanity();

	fprintf(stderr, "-- Final sanity check\n");
	sanity();
#endif

	dump_schema();
	dump_instances();
	dump_edges();
	dump_attributes();
}

static int			read_object_suffix;
static int			read_object_seen_section;
static Csource		*read_object_sourceP;
static Csource		*read_object_executableP;
static Centity		*read_object_source_underP;
static Carchive		*read_object_archiveP;
static Cfunction	*read_object_functionP;
static Cfile		*read_object_fileP;
static int			read_object_lineno;
static int			read_object_seen_end;


static void
read_object_callback(int lth, char *inputP)
{
	Cfunction	*callingP;
	Cedge		*edgeP;
	char		*P, *P1, *P2, *absolute_pathP;
	int			c;

	if (!lth) {
		goto done;
	}
	switch (c = *inputP) {
	case 'A':
		if (!strncmp(inputP, "Archive ", 8)) {
			P              = inputP + 8;
			absolute_pathP = absolute_path(P);
			read_object_archiveP = new Carchive(absolute_pathP);
			if (!read_object_archiveP) {
				outofmemory();
			}
			read_object_source_underP = read_object_archiveP;
		}
		goto done;
	case 'F':
		if (!strncmp(inputP, "File ", 5)) {
			P    = inputP + 5;
			lth -= 5;
			read_object_sourceP = new Csource(P, read_object_archiveP);
			if (!read_object_sourceP) {
				outofmemory();
			}
			goto done;
		}
		if (!strncmp(inputP, "Function ", 9)) {
			P = inputP + 9;
			goto is_function;
		}
		goto done;
	case 'C':
		if (!strncmp(inputP, "Calls ", 6)) {
			P = inputP + 6;
			goto is_call;
		}
		goto done;
	case 'D':
		if (!read_object_seen_section) {
			if (!strncmp(inputP, "Disassembly of section ", 23)) {
				if (!strcmp(inputP+23, ".init")) {
					/* This is an object or .so file */
					read_object_executableP = read_object_sourceP;
			}	}
			read_object_seen_section = 1;
		}
		goto done;
	case 'S':
		if (!strncmp(inputP, "Source ", 7)) {
			P              = inputP + 7;
			absolute_pathP = 0;
			if (read_object_executableP) {
				absolute_pathP = absolute_path(P);
				read_object_sourceP = Csource::locate(absolute_pathP, read_object_executableP);
			}
			read_object_fileP = Cfile::locate(read_object_sourceP, 0);
			if (!read_object_fileP) {
				if (!absolute_pathP) {
					absolute_pathP = absolute_path(P);
				}
				read_object_fileP = new Cfile(read_object_sourceP, 0, absolute_pathP);
				if (!read_object_fileP) {
					outofmemory();
			}	}
			read_object_lineno = 0;
		}
		goto done;
	case 'L':
		if (!strncmp(inputP, "Line ", 5)) {
			P = inputP + 5;
			read_object_lineno = strtol(P, &P1, 10);
		}
		goto done;
	case 'E':
		if (!strncmp(inputP, "End of report", 13)) {
			read_object_seen_end = 1;
		}
		goto done;
	}
	if (lth < 10) {
		goto done;
	}
	if (lth > 28 && inputP[8] == ':') {
		P = inputP + 32;
		if (!strncmp(P, "call ", 5)) {
			P = strchr(P, '<');
			if (!P) {
				goto done;
			}
			++P;
			P1 = strchr(P, '>');
			if (!P1) {
				goto done;
			}
			*P1 = 0;
			if (P1-P > 4) {
				P2 = P1 - 4;
				if (!strcmp(P2, "@PLT")) {
					*P2 = 0;
			}	}
is_call:
			callingP = Cfunction::locate(P, FindX, read_object_sourceP);
			if (!read_object_functionP) {
				/* Don't know how this could happen */
				goto done;
			}
			edgeP = Cedge::locate(read_object_functionP, callingP, read_object_lineno, read_object_fileP);
			edgeP->m_freq++;
		}
		goto done;
	}

	if ((c >= '0' && c <= '9') || (c >= 'a' && c <= 'f')) {

		P = inputP + 9;
		if (*P++ != '<') {
			goto done;
		}
		P1 = strchr(P, '>');
		if (!P1) {
			goto done;
		}
		*P1 = 0;
is_function:
		read_object_functionP = Cfunction::locate(P, FindX, read_object_sourceP);
		if (!read_object_functionP->m_fileP) {
			read_object_functionP->m_fileP  = read_object_fileP;
			read_object_functionP->m_lineno = read_object_lineno;
		} 
		goto done;
	}
done:
	return;
}

static void
read_object_init(int suffix)
{
	read_object_suffix      = suffix;
	read_object_seen_section= 0;
	read_object_archiveP    = 0;
	read_object_executableP = 0;
	read_object_source_underP = 0;
	read_object_sourceP     = 0;
	read_object_fileP       = 0;
	read_object_lineno      = 0;
	read_object_functionP   = 0;
	read_object_seen_end    = 0;
}

static void
read_object_final(char *filenameP)
{
	if (!read_object_seen_end)  {
		fprintf(stderr, "** Line %d: Failed to fully process %s\n", lineno, filenameP);
		g_ret = 1;
}	}

static void
read_object(char *filenameP, int suffix)
{
	read_object_init(suffix);
	objdump_init(g_program_nameP);
	objdumpfile(filenameP, read_object_callback);
	objdump_finalize();
	read_object_final(filenameP);
}

static void
read_cx(char *filenameP, int suffix, char *argv[])
{
	read_object_init(suffix);
	cx(argv, read_object_callback);
	read_object_final(filenameP);
}

static char *
end_symbol_name(char *P)
{
	char	*P1;
	int		c;

	c = *P;
	if (c == '_' || isalpha(c)) {
		for (P1 = P; function_char[(int) *P1]; ++P1);
		return(P1);
	}
	return(0);
}

int		g_file_memory_size = 0;
char	*g_file_memoryP    = 0;


static void
read_asm_pass1(Csource *sourceP)
{
	char		*atP, *nextP;

	Cfunction	*functionP;
	Cfile		*fileP;
	Cvariable	*usesP, *uses_nextP;
	char		*P, *P1, *P2, *absolute_pathP;
	int			number;
	int			lno;
	int			seen_global, seen_local, seen_weak, flags, label;
	char		c, quote;
		
	fileP       = 0;
	lno         = 0;
	functionP   = 0;
	seen_global = 0;
	seen_local  = 0;
	seen_weak   = 0;

	for (atP = g_file_memoryP; ;*nextP = '\n', atP = nextP + 1) {
		quote        = 0;
		label        = 1;
		for (nextP = atP; ; ++nextP) {
			c = *nextP;
			if (label) {
				if (label_char[(int) c]) {
					continue;
				}
				if (c == ':') {
					atP = ++nextP;
					c   = *nextP;
				}
				label = 0;
			}

			switch (c) {
			case 0:
				return;
			case '\\':
				++nextP;
				continue;
			case '\t':
				if (!quote) {
					*nextP = ' ';
				}
				continue;
			case ';':
				if (quote) {
					continue;
				}
			case '\n':
				break;
			case '#':
				if (quote) {
					continue;
				}
				for (P = nextP; *P != '\n'; *P++ = ' ');
				nextP = P-1;
				continue;
			case '/':
				if (quote) {
					continue;
				}
				if (nextP[1] == '*') {
					*nextP   = ' ';
					P        = nextP + 1;
					*P++     = ' ';
					for (; ;) {
						switch (*P) {
						case 0:
							P[-1] = '\n';
							nextP = P - 2;
							break;
						case '*':
							if (P[1] == '/') {
								*P     = ' ';
								nextP  = P + 1;
								*nextP = ' ';
								break;
							}
						default:
							*P = ' ';
							continue;
						}
						break;
				}	}
				continue;
			case '"':
			case '\'':
				if (!quote) {
					quote = c;
				} else if (quote == c) {
					quote = 0;
				}
			default:
				continue;
			}
			break;
		}
		*nextP = 0;
		for (P = atP; *P == ' '; ++P);
		switch (*P) {
		case 0:
			continue;
		case '.':
			++P;
			switch (*P) {
			case 'g':
				if (!strncmp(P, "globl ", 6)) {		// .globl
					seen_global = 1;
				}
				continue;
			case 'w':
				if (!strncmp(P, "weak ", 5)) {		// .weak
					seen_weak = 1;
				}
				continue;
			case 'c':
				if (!strncmp(P, "comm ", 5)) {		// .comm
					flags = 0;
					if (seen_local) {
						seen_local = 0;
						flags      = LocalX;
					}
					if (show_variables != No_variables) {
						for (P += 5; *P == ' '; ++P);
						P1 = end_symbol_name(P);
						if (P1) {
							c     = *P1;
							*P1   = 0;
							usesP = Cvariable::locate(P, flags, sourceP, functionP);
							*P1   = c;
				}	}	}
				continue;
			case 't':
				if (!strncmp(P, "type ", 5)) {	
					flags = LocalX;
					if (seen_weak) {
						flags       = WeakX;
						seen_weak   = 0;
					}
					if (seen_global) {
						flags       = 0;
						seen_global = 0;
					}
					for (P += 5; *P == ' '; ++P);
					P1 = end_symbol_name(P);
					if (!P1) {
						continue;
					}
					P2 = strchr(P1, ',');
					if (!P2) {
						continue;
					}
					for (++P2; *P2 == ' '; ++P2);
					if (!strncmp(P2, "@object", 7)) {
						if (show_variables != No_variables) {
							c   = *P1;
							*P1 = 0;
							flags |= InitX;
							usesP = Cvariable::locate(P, flags, sourceP, functionP);
							*P1 = c;
						}
						continue;
					}
					if (!strncmp(P2, "@function", 9)) {
						c   = *P1;
						*P1 = 0;
						functionP = Cfunction::locate(P, flags, sourceP);
						functionP->m_flags |= flags;
						*P1 = c;
						// Nested variables are declared before the
						// function that contains them
						for (usesP = Cvariable::g_local_headP; usesP; usesP = uses_nextP) {
							uses_nextP           = usesP->m_local_nextP;
							usesP->m_parentP     = functionP;
							usesP->m_local_nextP = 0;
						}
						Cvariable::g_local_headP = 0;
					}
				}
				continue;
			case 'f':
				if (!strncmp(P, "file ", 5)) {
					for (P += 5; *P == ' '; ++P);
					number = strtol(P, &P1, 10);
					if (P == P1) {
						continue;
					}
					if (*P1 != ' ') {
						continue;
					}
					for (++P1; *P1 == ' '; ++P1);
					if (*P1 != '"') {
						continue;
					}
					P = P1 + 1;
					P1 = strchr(P, '"');
					if (!P1) {
						continue;
					}
					c   = *P1;
					*P1 = 0;
					absolute_pathP = absolute_path(P);
					*P1 = c;
					fileP = new Cfile(sourceP, number, absolute_pathP);
					if (!fileP) {
						outofmemory();
					}
				}
				continue;
			case 'l':
				if (!strncmp(P, "loc", 3)) {
					P += 3;
					if (*P == ' ') {	// .loc
						for (; *P == ' '; ++P);
						number = strtol(P, &P1, 10);
						if (P == P1) {
							continue;
						}
						fileP = Cfile::locate(sourceP, number);
						if (!fileP) {
							// This should never happen
							continue;
						}
						for (P = P1; *P == ' '; ++P);
						lno = strtol(P, &P1, 10);
						if (P == P1) {
							continue;
						}
						if (functionP && !functionP->m_fileP) {
							functionP->m_fileP  = fileP;
							functionP->m_lineno = lno;
						}
						continue;
					}
					if (!strncmp(P, "al ", 3)) {	// .local
						seen_local = 1;
				}	}
			}
			continue;
}	}	}

/*
 http://www.kernel-traffic.org/wine/wn20000228_32.html

 The linker/loader maintain a global offset table, which consists of an array
 of pointers to exported variables. All access to those variables is supposed
 to go through these pointers, if the variable could reside within another
 dynamic object. 

 If 'variable' is an exported symbol, then 'variable@GOT' is replaced by the
 linker with a constant: the offset of the GOT slot containing the pointer to
 variable. So, if the %ebx register contains the start address of the GOT, you
 can retrieve this pointer via 

 movl variable@GOT(%ebx), %eax 

 If you want the value of the variable, you need another indirection: 

 movl (%eax), %eax 

 If you know, however, that the variable in question must reside in the same
 dynamic object as the user, then you can avoid that double indirection, by
 using the GOTOFF facility. 'variable@GOTOFF' is replaced by the linker with
 another constant: the difference between the absolute address of the variable
 and the address of the global offset table (of the current dynamic object).
 Thus, you can retrieve the value of the variable in one step using: 

 movl variable@GOTOFF(%ebx), %eax 

*/

static void
read_asm_pass2(Csource *sourceP)
{
	char		*atP, *nextP;

	Cfunction	*functionP;
	Cfile		*fileP;
	Cfunction	*callingP;
	Cclassmember *usesP;
	Cedge		*edgeP;
	char		*P, *P1, *P2;
	int			number;
	int			lno;
	int			bracket, update1, update2, flags, label;
	char		c, c2, c3;

	fileP       = 0;
	lno         = 0;
	functionP   = 0;
	for (atP = g_file_memoryP; *atP; *nextP = '\n', atP = nextP + 1) {

		label        = 1;
		for (nextP = atP; ; ++nextP) {
			c = *nextP;
			if (label) {
				if (label_char[(int) c]) {
					continue;
				}
				if (c == ':') {
					atP = nextP + 1;
				}
				label = 0;
			}
			break;
		}
		nextP = strchr(atP, '\n');
		if (!nextP) {
			return;
		}
		*nextP = 0;
		for (P = atP; *P == ' '; ++P);
		switch (*P) {
		case 0:
			continue;
		case '.':
			++P;		
			switch (*P) {
			case 't':
				if (!strncmp(P, "type ", 5)) {		// .type
					for (P += 5; *P == ' '; ++P);
					P1 = end_symbol_name(P);
					if (!P1) {
						continue;
					}
					P2 = strchr(P1, ',');
					if (!P2) {
						continue;
					}
					for (++P2; *P2 == ' '; ++P2);
					if (!strncmp(P2, "@function", 9)) {
						c   = *P1;
						*P1 = 0;
						functionP = Cfunction::locate(P, 0, sourceP);
						*P1 = c;
				}	}
				continue;
			case 'f':
				if (!strncmp(P, "file ", 5)) {		// .file
					for (P += 5; *P == ' '; ++P);
					number = strtol(P, &P1, 10);
					if (P == P1) {
						continue;
					}
					if (*P1 != ' ') {
						continue;
					}
					fileP = Cfile::locate(sourceP, number);
				}
				continue;
			case 'l':
				if (!strncmp(P, "loc", 3)) {		// .loc
					P += 3;
					if (*P == ' ') {	// .loc
						for (; *P == ' '; ++P);
						number = strtol(P, &P1, 10);
						if (P == P1) {
							continue;
						}
						fileP = Cfile::locate(sourceP, number);
						if (!fileP) {
							// This should never happen
							continue;
						}
						for (P = P1; *P == ' '; ++P);
						lno = strtol(P, &P1, 10);
						continue;
				}	}
			}
			continue;
		case 'c':
			if (no_function_calls) {
				continue;
			}
			if (!strncmp(P, "call ", 5)) {		//\tcall
				for (P  += 5; *P == ' '; ++P);
				P1  = P;
				if (*P1 == '*') {
					++P1;
				}
				if (*P1 == '%') {
					++P1;
				}
				P1  = end_symbol_name(P1);
				if (!P1) {
					continue;
				}
				c   = *P1;
				*P1 = 0;
				callingP = Cfunction::locate(P, FindX, sourceP);
				*P1 = c;
				if (functionP) {
					edgeP = Cedge::locate(functionP, callingP, lno, fileP);
					edgeP->m_freq++;
				}
				continue;
			}
		default:
			if (no_function_calls && show_variables == No_variables) {
				continue;
			}
			c = *P;
			if (c < 'a') {
				c += 'a' - 'A';
			}
			if (c < 'a' || c > 'z') {
				continue;
			}
			c2 = P[1];
			if (c2 < 'a') {
				c2 += 'a' - 'A';
			}
			c3 = P[2];
			if (c3 < 'a') {
				c3 += 'a' - 'A';
			}
			
			update1 = 0;
			update2 = UpdateX;
			switch (c) {
			case 'b':
				switch (c2) {
				case 't':						// bt
					if (c3 == 'c' ||			// btc
					    c3 == 'r' ||			// btr
						c3 == 's') {			// bts
						break;
					}
					update2 = 0;
					break;
				case 'o':
					if (c3 == 'u') {			// bound
						update2 = 0;
					}
					break;
				}
				break;
			case 'c':
				if (c2 == 'm' && c3 == 'p') {	// cmp
					if (P[3] != 'x' && P[3] != 'X') {	// cmpx
						update2 = 0;
				}	}
				break;
			case 'd':
				if (c2 == 'e' && c3 == 'c') {	// dec
					update1 = UpdateX;
				}
				break;
			case 'e':
				if (c2 == 'n' && c3 == 't') {	// enter
					update2 = 0;
				}
				break;
			case 'i':
				if (c2 == 'n' && c3 == 'c') {	// inc
					update1 = UpdateX;
				}
				break;
			case 'j':							// jmp etc.
				continue;
			case 'l':
				if (c2 == 'o' && c3 == 'o') {	// loop
					continue;
				}
				break;
			case 'n':
				switch (c2) {
				case 'e':
					if (c3 == 'g') {			// neg
						update1 = UpdateX;
					}
					break;
				case 'o':
					if (c3 == 't') {			// not
						update1 = UpdateX;
					}
					break;
				}
				break;
			case 'o':
				if (c2 == 'u' && c3 == 't') { 	// out
					continue;
				}
				break;
			case 'p':
				if (c2 == 'o' && c3 == 'p') {	// pop
					update1 = UpdateX;
				}
				break;
			case 's':
				if (c2 == 'e' && c3 == 't') {	// set
					update1 = UpdateX;
				}
				break;
			case 't':
				if (c2 == 'e' && c3 == 's') {	// test
					update2 = 0;
				}
				break;
			case 'x':
				if (c2 == 'c' && c3 == 'h') {	// xchg
					update1 = UpdateX;
				}
				break;
			}

			P = strchr(P, ' ');
			if (!P) {
				continue;
			}
			for (++P; *P == ' '; ++P);
			if (*P == '$') {	// Address of
				flags = AddressesX | FindX;
				update1 = UpdateX;
				++P;
			} else {
				flags = FindX;
				if (show_variables == No_variables) {
					continue;
			}	}
			P1 = end_symbol_name(P);
			if (P1) {
				c     = *P1;
				*P1   = 0;
				// Variable anywhere
				usesP = Cvariable::locate(P, flags, sourceP, functionP);
				if (functionP) {
					edgeP = Cedge::locate(functionP, usesP, lno, fileP);
					edgeP->m_freq++;
					edgeP->m_flags |= update1;
					usesP->m_flags |= UsedX;
				}
				*P1 = c;
				P   = P1;
			}
			// Search for ','
			bracket = 0;
			for (;; ++P) {
				switch (*P) {
				case ',':
					if (bracket) {
						// Ignore comma in brackets
						continue;
					}
				case 0:
					break;
				case '(':
					++bracket;
					continue;
				case ')':
					--bracket;
				default:
					continue;
				}
				break;
			}

			if (*P != ',') {
				continue;
			}
			for (; (c = *++P) && c == ' '; );
			if (c == '$') {
				flags = AddressesX | FindX;
				update2 = UpdateX;
				c = *++P;
			} else {
				flags = FindX;
			}
			P1 = end_symbol_name(P);
			if (P1) {
				*P1 = 0;
				usesP = Cvariable::locate(P, flags, sourceP, functionP);
				if (functionP) {
					edgeP = Cedge::locate(functionP, usesP, lno, fileP);
					edgeP->m_freq++;
					edgeP->m_flags |= update2;
					usesP->m_flags |= UsedX;
	}	}	}	}
	return;
}

#ifndef SSIZE_MAX
#define SSIZE_MAX 0x7FFFFFFF
#endif

static void
read_asm(int fileno, Csource *sourceP)
{
	struct stat buf;
	char		*atP, *endP;
	int			lth, segment;
	int			ret;

	// We have to scan the input more than once so read the entire input
	// into memory

	ret = fstat(fileno, &buf);
	if (ret != 0) {
		fprintf(stderr,"** Line %d: fstat %s returned %d\n", lineno, sourceP, ret);
		perror("");
		return;
	}

	// Allocate a large enough memory buffer

	lth = buf.st_size;
	if (!lth) {
		fprintf(stderr, "** Line %d: %s is empty\n", lineno, sourceP);
		return;
	}

	if (lth >= g_file_memory_size) {
		free(g_file_memoryP);
		g_file_memory_size = ((lth + 2) + 2047) & ~2047;
		g_file_memoryP = (char *) malloc(g_file_memory_size);
		if (!g_file_memoryP) {
			outofmemory();
	}	}

	atP = g_file_memoryP;
	for (endP = atP + lth; atP < endP; atP += ret) {
		segment = endP - atP;
		if (segment > SSIZE_MAX) {
			segment = SSIZE_MAX;
		}
		ret = read(fileno, atP, segment);
		if (ret <= 0) {
			fprintf(stderr, "** Line %d: read %s returned %d\n", lineno, sourceP, ret);
			perror("");
			return;
	}	}
	assert(atP == endP);
	if (atP[-1] != '\n') {
		*endP++ = '\n';
	}
	*endP = 0;
		
	read_asm_pass1(sourceP);
	read_asm_pass2(sourceP);
}

static void
get_environment(char **argv)
{
	int		i, j, c, suffix;
	char	*P, *P1, *P2, *suffixP;

	for (i = 'A'; i <= 'Z'; ++i) {
		label_char[i]    = 1;
		function_char[i] = 1;
	}
	for (i = 'a'; i <= 'z'; ++i) {
		label_char[i]    = 2;
		function_char[i] = 2;
	}
	for (i = '0'; i <= '9'; ++i) {
		label_char[i]    = 3;
		function_char[i] = 3;
	}
	label_char[(int) '_']    = 4;
	function_char[(int) '_'] = 4;
	function_char[(int) '.'] = 5;
	function_char[(int) '$'] = 6;

	for (i = j = 1; P = argv[i]; ++i) {
		if (P[0] == '-' && P[1] == '-') {
			if (!strncmp(P, "--ASX_COMPILE=", 14)) {
				asx_compileP = P + 14;
				continue;
			}
			if (!strncmp(P, "--ASX_UNLINK=", 13)) {
				asx_unlinkP = P + 13;
				continue;
			}
			if (!strncmp(P, "--ASX_IGNORE=", 13)) {
				asx_ignoreP = P + 13;
				continue;
			}
			if (!strncmp(P, "--ASX_FORCE=", 12)) {
				asx_forceP = P + 12;
				continue;
			}
			if (!strncmp(P, "--ASX_SILENT=", 13)) {
				asx_silentP = P + 13;
				continue;
		}	}
		argv[j++] = P;
	}
	argv[j] = 0;


	if (!asx_compileP) {
		asx_compileP = getenv("ASX_COMPILE");
	}
	if (!asx_silentP) {
		asx_silentP = getenv("ASX_SILENT");
	}
	if (!asx_ignoreP) {
		asx_ignoreP = getenv("ASX_IGNORE");
	}
	if (!asx_forceP) {
		asx_forceP = getenv("ASX_FORCE");
	}

	if (asx_compileP) {
		printf("// ASX_COMPILE=%s// Will not compile source code", asx_compileP);
		if (!strcasecmp(asx_compileP, "no")) {
			printf(" (Extract by directly examining source code)\n");
			compile_flag = 0;
			unlink_flag  = 0;
		} else {
			printf(" (Extract by compiling source code to assembler)\n");
	}	}
		
	if (asx_silentP) {
		printf("// ASX_SILENT=%s// Will not echo commands", asx_silentP);
		if (!strcasecmp(asx_silentP, "no")) {
			printf(" (Echoing compiler commands)\n");
			silent_flag = 0;
		} else {
			printf(" (Compiler commands not shown)\n");
			silent_flag = 1;
	}	}
		
	if (compile_flag) {
		if (!asx_unlinkP) {
			asx_unlinkP = getenv("ASX_UNLINK");
		}
		if (asx_unlinkP) {
			printf("// ASX_UNLINK=%s", asx_unlinkP);
			if (!strcasecmp(asx_unlinkP, "no")) {
				unlink_flag = 0;
				printf(" (Generated *.s files preserved)\n");
			} else {
				printf(" (Generated *.s files removed)\n");
	}	}	}

	if (asx_ignoreP) {
		printf("// ASX_IGNORE=%s\n", asx_ignoreP);

		for (P = asx_ignoreP; P; P = (c ? P1+1 : 0)) {
			P1 = P;
			c  = *P;
			// Remove leading space
			if (isspace(c)) {
				continue;
			}
			for (; (c = *P1) && c != ';' && c != ':' && c != ','; ++P1);
			*P1 = 0;
			// Remove trailing space
			for (P2 = P1; --P2 >= P; ) {
				if (!isspace(*P2)) {
					break;
				}
				*P2 = 0;
			}
			switch (*P) {
			case 'f':
			case 'F':
				no_function_calls = 1;
				continue;
			case 'v':
			case 'V':
				switch (P[1]) {
				case 's':
				case 'S':
					show_variables = Global_variables;
					continue;
				case 'f':
				case 'F':
					show_variables = Static_variables;
					continue;
				}
				show_variables = No_variables;
				continue;
			case 0:
				// Ignore things that only contained whitespace
				continue;
			}

			for (suffix = 0; ; ++suffix) {
				suffixP = suffixes[suffix];
				if (!suffixP) {
					fprintf(stderr, "** Can't ignore suffix \"%s\" - unknown\n", P);
					break;
				}
				if (!strcmp(P, suffixP)) {
					ignore[suffix] = 1;
					break;
			}	}
	}	}

	if (asx_forceP) {
		printf("// ASX_FORCE=%s\n", asx_forceP);
		for (P = asx_forceP; P; P = (c ? P1+1 : 0)) {
			P1 = P;
			c  = *P;
			// Remove leading space
			if (isspace(c)) {
				continue;
			}
			for (; (c = *P1) && c != ';' && c != ':' && c != ','; ++P1);
			*P1 = 0;
			// Remove trailing space
			for (P2 = P1; --P2 >= P; ) {
				if (!isspace(*P2)) {
					break;
				}
				*P2 = 0;
			}
			switch (*P) {
			case 'v':
			case 'V':
				switch(P[1]) {
				case 'g':
				case 'G':
					show_variables = Global_variables;
					continue;
				case 'S':
					show_variables = Static_variables;
					continue;
				}
				show_variables = Static_function_variables;
				continue;
			case 0:
				// Ignore things that only contained whitespace
				continue;
			}

			for (suffix = 0; ; ++suffix) {
				suffixP = suffixes[suffix];
				if (!suffixP) {
					fprintf(stderr, "** Can't force suffix \"%s\" - unknown\n", P);
					break;
				}
				if (!strcmp(P, suffixP)) {
					force[suffix] = 1;
					break;
	}	}	}	}
	return;
}

static int
parse_command(char *commandP, char **argv1P)
{
	char	*startP, *endP, *nextP;
	int		i, c;

	startP = commandP;
	if (*startP != '"') {
		fprintf(stderr, "** Line %d: Missing cwd start '\"' >%s\n", lineno, startP);
		return(-1);
	}

	for (i = -1; startP; ++i) {
		for (endP = nextP = ++startP; (c = *nextP) != '"'; ++nextP) {
			switch (c) {
			case 0:
				fprintf(stderr, "** Line %d: Missing end '\" >%s'\n", lineno, startP);
				return(-1);
			case '\\':
				c = *(++nextP);
			default:
				*endP++ = c;
		}	}
		*endP = 0;
		if (i < 0) {
			cwdP = startP;
			if (chdir(cwdP)) {
				fprintf(stderr,"** Line %d: Can't change directory to %s\n", lineno, cwdP);
				perror("");
				return(-1);
			}
		} else {
			argv1P[i] = startP;
		}
		startP = strchr(nextP+1, '"');
	}
	if (i < 1) {
		fprintf(stderr, "** Line %d: Missing cmd '\" >%s'\n", lineno, commandP);
		return(-1);
	}
	argv1P[i]   = 0;
	return(i);
}

static int
process_command(int argc2, char **argv1P, char * const env[])
{
	static char	filename1[1024];
	static char	*compilePP[2048];
	int			compile;

	char 	*programP, *compileP, *targetP, *assemblerP, *absolute_pathP;
	char	*P, *P1, *suffixP;
	int		i, suffix, unlink_me, lth, ret, fno, argc1;
	int		seen_dash_S, seen_dash_E, running_as;
	int		c2;
	Csource	*sourceP;

	/* Find the source and target files */

	seen_dash_S = 0;
	seen_dash_E = 0;
	targetP     = 0;
	assemblerP  = 0;
	compile     = 0;
	running_as  = 0;
	programP    = argv1P[0];
	lth         = strlen(programP);
	if (lth > 3) {
		if (!strcmp(programP+lth-3, "/as")) {
			running_as = 1;
		} else if (lth > 4 && !strcmp(programP+lth-4, "/gas")) {
			running_as = 1;
	}	}
	
	if (running_as) {
		for (argc1 = i = 1; P = argv1P[argc1]; ++argc1) {
			if (argc1 >= argc2) {
				fprintf(stderr, "** lineno %d: Can't scan args for %s\n", lineno, programP);
				return(-1);
			}
			if (*P != '-') {	
				if (!strcmp(P, "/dev/null")) {
					return(0);
				}
				compilePP[compile++] = P;
				continue;		// Discard (to add back in later)
			}
			switch(P[1]) {
			case '-':
				if (!strcmp(P,"--defsym")) {
					++argc1;	// Discard two arguments
				}
				continue;
			case 'I':
			case 'o':
				++argc1;
			case 'a':
			case 'D':
			case 'f':
			case 'g':
			case 'J':
			case 'K':
			case 'L':
			case 'n':
			case 'R':
			case 'v':
			case 'W':
			case 'w':
			case 'x':
			case 'Z':
				continue;
			}
			fprintf(stderr, "** Line %d: Unknown %s option '%s'\n", lineno, programP, P);
			goto bad;
		}
	} else {
		for (argc1 = i = 1; P = argv1P[argc1]; ++argc1) {
			if (argc1 >= argc2) {
				fprintf(stderr, "** lineno %d: Can't scan args for %s\n", lineno, programP);
				return(-1);
			}
			if (*P != '-') {	
				if (!strcmp(P, "/dev/null")) {
					return(0);
				}
				compilePP[compile++] = P;
				continue;		// Discard (to add back in later)
			}
			c2 = P[2];
			switch(P[1]) {
			case 'a':
				if (!strcmp(P, "-aux-info")) { 		// Discard 2 args
					++argc1;
					continue;
				}
			case 'B':
			case 'f':
			case 'n':
			case 't':
			case 'Y':
				argv1P[i++] = P;
				continue;
			case 'D':
			case 'I':
			case 'U':
				if (!c2) {
					argv1P[i++] = P;
					P           = argv1P[++argc1];
				}
				argv1P[i++] = P;
				continue;
			case 'b':
				if (!c2) {
					++argc1;						// Discard 2 args
					continue;
				}
				argv1P[i++] = P;
			case 'c':
			case 'C':
			case 'g':
			case 'h':
			case '-':
			case 'M':
			case 'O':
			case 'P':
			case 'Q':
			case 'r':
			case 'v':
			case 'w':
			case 'l':
			case 'L':
				continue;							// Discard
			case 'd':
				if (c2 == 'D' || c2 == 'I' || c2 == 'M' || c2 == 'N') {
					argv1P[i++] = P;
				}
				continue;
			case 'E':
				if (!c2) {
					seen_dash_E = 1;
					continue;
				}
				argv1P[i++] = P;
				continue;
			case 'G':
			case 'm':
				if (!c2) {
					++argc1;
				}
				continue;
			case 'i':
				if (!strcmp(P, "-idirafter")     ||
					!strcmp(P, "-imacros")       ||
					!strcmp(P, "-include")       ||
					!strcmp(P, "-iprefix")       ||
					!strcmp(P, "-iwithprefix")   ||
					!strcmp(P, "-isystem")) {
					argv1P[i++] = P;
					P           = argv1P[++argc1];
				}
				argv1P[i++] = P;
				continue;
			case 'o':	// <file>
				if (!c2) {
					targetP = argv1P[++argc1];	// Discard parameter
					continue;
				}
				break;
			case 'p':
				if (strcmp(P, "-print")) {		// Not -print
					argv1P[i++] = P;
				} else {
					if (!strcmp(P, "-print-file-name")) {
						return(0);
				}	}
				continue;
			case 'S':
				if (!c2) {
					seen_dash_S = 1;
					continue;	
				}
				break;
			case 's':
				if (!strcmp(P, "-std")) {
					argv1P[i++] = P;
				}
				continue;
			case 'u':
			case 'V':
				if (!c2) {
					argv1P[i++] = P;
					P           = argv1P[++argc1];
				}
				argv1P[i++] = P;
				continue;
			case 'W':
				if (!strcmp(P, "-Wp,")) {
					if (!strcmp(P, "-Wp,-M")) {
						argv1P[i++] = P;
				}	}
				continue;
			case 'x':
				if (c2) {
					argv1P[i++] = P;
					continue;
				}
			case 'X':
				argv1P[i++] = P;
				argv1P[i++] = argv1P[++argc1];
				continue;
			case '#':
				if (!strcmp(P, "-###")) {
					continue;
				}
				break;
			}
			/* If the compiler was not compiling we shouldn't */
			fprintf(stderr, "** Line %d: Unknown compiler option '%s'\n", lineno, P);
			goto bad;
		}
	}

	argc1 = i;

	compilePP[compile] = 0;

	if (seen_dash_S) {
		assemblerP = targetP;
	}

	for (compile = 0; compileP = compilePP[compile]; ++compile) {
		P1 = strrchr(compileP, '.');
		if (!P1) {
			suffix = SUFFIX_none;
		} else {
			++P1;
			for (suffix= 0;; ++suffix) {
			
				if (!(suffixP = suffixes[suffix])) {
					fprintf(stderr, "** Line %d: Unknown suffix %s\n", lineno, compileP);
					goto next_command;
				}
				if (!strcmp(P1, suffixP)) {
					break;
		}	}	}
	
		if (ignore[suffix]) {
			goto next_command;
		}

		unlink_me  = 0;
		if (suffix == SUFFIX_S && running_as) {
			suffix = SUFFIX_s;
		}
		switch (suffix) {
		case SUFFIX_so:
		case SUFFIX_o:
			if (use_stdin && !force[suffix]) {
				int		s;
				char	*suffixP;
				
				lth = P1 - compileP;
				strncpy(filename1, compileP, lth);
	
				if (suffix == SUFFIX_so) {
					s = SUFFIX_o;
				} else {
					s = SUFFIX_s;
				}
				for (; s < SUFFIX_none; ++s) {
					suffixP = suffixes[s];
					if (!suffixP) {
						continue;
					}
					strcpy(filename1+lth, suffixP);
					absolute_pathP = absolute_path(filename1);
					if (Csource::lookup(absolute_pathP)) {
						fprintf(stderr, "-- Line %d: Assuming %s generated from %s\n", lineno, compileP, absolute_pathP);
						goto next_command;
			}	}	}
		case SUFFIX_none:
		case SUFFIX_a:
			read_object(compileP, suffix);
			break;
		case SUFFIX_cxx:
		case SUFFIX_cpp:
		case SUFFIX_cpp1:
			if (!use_stdin) {
				argv1P[0] = "/usr/bin/g++";
			}
		case SUFFIX_c:
		case SUFFIX_C:
		case SUFFIX_cc:
			if (seen_dash_E) {
				goto next_command;
			}
		case SUFFIX_S:
			if (!assemblerP) {
				lth = P1-compileP;
				memcpy(filename1, compileP, lth);
				strcpy(filename1+lth, "s");
				assemblerP = filename1;
			}
			if (!compile_flag) {
				goto read_assembler;
			}
			
			argv1P[argc1++] = compileP;
			argv1P[argc1]   = 0;
			argv1P         -= 4;
			argv1P[0]       = argv1P[4];
			argv1P[1]       = "-o";
			argv1P[2]       = assemblerP;
			argv1P[3]       = "-g";
			if (suffix == SUFFIX_S) {
				argv1P[4]   = "-E";
			} else {
				argv1P[4]   = "-S";
			}
		
			if (unlink_flag && access(assemblerP, F_OK)) {
				unlink_me = 1;
			}
	
			ret = call_cc(argv1P, env, suffix);
			if (ret) {
				fprintf(stderr, "** Line %d: Failed to compile %s\n", lineno, assemblerP);
				argv1P[argc1-1] = absolute_path(assemblerP);
				read_cx(assemblerP, suffix, argv1P);
				unlink_me = 0;
				goto next_command;
			}
			goto read_assembler;
	
		case SUFFIX_s:
			assemblerP     = compileP;
read_assembler:
			absolute_pathP = absolute_path(compileP);
			sourceP        = new Csource(absolute_pathP, 0);
			if (!sourceP) {
				outofmemory();
				goto bad;
			}
		
			fno = open(assemblerP, O_RDONLY);
			if (fno < 0) {
				fprintf(stderr, "Line %d: Can't open %s\n", lineno, assemblerP);
				perror("");
				goto next_command;
			}
			read_asm(fno, sourceP);
			close(fno);
	
			if (!unlink_me) {
				break;
			}
	
			unlink_me = 0;
	
			/* Sanity checks follow */
			if (assemblerP != filename1) {
				fprintf(stderr, "Line %d: May not unlink file %s\n", lineno, assemblerP);
				exit(5);
				goto bad;
			} 
			lth = strlen(assemblerP);
			if (lth < 3 || assemblerP[lth-2] != '.' || assemblerP[lth-1] != 's') {
				fprintf(stderr, "Line %d: May not delete %s\n", lineno, assemblerP);
				exit(5);
				goto bad;
			}
	
			/* Discard this assembly file when done */
	
			if (unlink(assemblerP)) {
				fprintf(stderr, "Line %d: Can't delete %s\n", lineno, assemblerP);
				perror("");
				break;
			}
			break;
		}
next_command:
		continue;
	}
	return(0);
bad:
	return(-1);
}

// Leave room for -g -S -o filename
#define ExtraSlots 4

int
main(int argc, char *argv[], char * const env[])
{
	static char	*argv1[2048];
	
	char		**argv1P;
	int 		i;
	char		*stdin_bufferP;
	size_t		stdin_buffer_lth;
	clock_t		start_clock, end_clock;
	
	start_clock = time(0);

	printf("// ASX 2.0.1 Compiled " __DATE__ ":"  __TIME__ " Start: %s\n", ctime(&start_clock));

	get_environment(argv);

	printf("// Arguments to %s:\n//", argv[0]);

	for (i = 1; argv[i]; ++i) {
		fputc(' ', stdout);
		fputs(argv[i], stdout);
	}
	fputc('\n', stdout);

	g_program_nameP = argv[0];
	g_ret           = 0;

	if (argc < 2 || !strcmp(argv[1], "-")) {
    	use_stdin        = 1;
		stdin_buffer_lth = 120;
		stdin_bufferP    = (char *) malloc(stdin_buffer_lth);
		if (!stdin_bufferP) {
			fprintf(stderr, "Can't malloc stdin buffer\n");
			exit(1);
		}
		while (getline(&stdin_bufferP, &stdin_buffer_lth, stdin) >= 0) {
			argv1P = argv1 + ExtraSlots;
			++lineno;
			i = parse_command(stdin_bufferP, argv1P);
			if (i < 1) {
				continue;
			}
			process_command(i, argv1P, env);
		}
	} else {
		cwdP      = getcwd(0, 0);
		argv1P    = argv1 + ExtraSlots;
		argv1P[0] = "/usr/bin/gcc";
		for (i = 1; i < argc; ++i) {
			argv1P[i] = argv[i];	
		}
		argv1P[i]   = 0;
		process_command(i, argv1P, env);
	}

	Cvariable::resolve(show_variables, no_function_calls);
	Cfunction::resolve();
	collapse_directories();
	dump_ta();
	end_clock = time(0);
	printf("\n// End: %s \n", ctime(&end_clock));
	return(g_ret);
}

